/*
 * SCIM-chewing -
 *	Intelligent Chinese Phonetic IM Engine for SCIM.
 *
 * Copyright (c) 2004, 2005, 2006, 2008
 *	SCIM-chewing Developers. See ChangeLog for details.
 *
 * See the file "COPYING" for information on usage and redistribution
 * of this file.
 */

#ifndef SCIM_CHEWING_CONFIG_ENTRY_H
#define SCIM_CHEWING_CONFIG_ENTRY_H

#define SCIM_CHEWING_SELECTION_KEYS_NUM \
    "/IMEngine/Chewing/SelectionKeysNum"

#define SCIM_CONFIG_IMENGINE_CHEWING_SELECTION_KEYS \
    "1234567890"

#define SCIM_CONFIG_IMENGINE_CHEWING_USER_SELECTION_KEYS \
    "/IMEngine/Chewing/SelectionKeys"

#define SCIM_CONFIG_IMENGINE_CHEWING_USER_KB_TYPE \
    "/IMEngine/Chewing/KeyboardType"

#define SCIM_CONFIG_IMENGINE_CHEWING_ADD_PHRASE_FORWARD \
    "/IMEngine/Chewing/AddPhraseForward"

#define SCIM_CONFIG_IMENGINE_CHEWING_SPACE_AS_SELECTION \
    "/IMEngine/Chewing/SpaceAsSelection"

#define SCIM_CONFIG_IMENGINE_CHEWING_ESC_CLEAN_ALL_BUFFER \
    "/IMEngine/Chewing/EscCleanAllBuffer"

#define SCIM_CONFIG_IMENGINE_CHEWING_TRIGGER_KEY \
    "/IMEngine/Chewing/TriggerKey"

#define SCIM_CONFIG_IMENGINE_CHEWING_CHI_ENG_KEY \
    "/IMEngine/Chewing/ChiEngKey"

#define SCIM_CONFIG_IMENGINE_CHEWING_CHI_ENG_MODE \
    "/IMEngine/Chewing/ChiEngMode"

#define SCIM_CONFIG_IMENGINE_CHEWING_USE_CAPSLOCK \
    "/IMEngine/Chewing/UseCapslock"

#define SCIM_CONFIG_IMENGINE_CHEWING_SHOW_CANDIDATE_COMMENT \
    "/IMEngine/Chewing/ShowCandidateComment"


#define SCIM_CONFIG_IMENGINE_CHEWING_PINYIN_METHD \
    "/IMEngine/Chewing/PinYinMethod"

#define SCIM_CONFIG_IMENGINE_CHEWING_EXTERNAL_PINYIN_PATH \
    "/IMEngine/Chewing/ExternalPinYinPath"

/* It's known to be at least 5. */
#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_NUM (5)

#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_ \
    "/IMEngine/Chewing/PreeditBackgroundColor"

#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_DEF_1 \
    "#A7A7A7"
#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_DEF_2 \
    "#C5C5C5"
#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_DEF_3 \
    "#A7A7A7"
#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_DEF_4 \
    "#F0F0F0"
#define SCIM_CONFIG_IMENGINE_CHEWING_PREEDIT_BGCOLOR_DEF_5 \
    "#AAAAAA"

#define SCIM_CONFIG_IMENGINE_CHEWING_PHRASE_CHOICE_REARWARD \
    "/IMEngine/Chewing/PhraseChoiceRearward"

#define SCIM_CONFIG_IMENGINE_CHEWING_AUTO_SHIFT_CURSOR \
    "/IMEngine/Chewing/AutoShiftCursor"

#endif
